//*-- Author:  Ole Hansen<mailto:ole@jlab.org>; Jefferson Lab; 11-Jan-2010
//
#ifndef ROOT_TreeSearch_GEM
#define ROOT_TreeSearch_GEM

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// TreeSearch::GEM                                                           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include "THaTrackingDetector.h"
#include "THaDetMap.h"
#include "Types.h"
#include "TMatrixDSym.h"
#include <vector>
#include <utility>
#include <set>
#include <list>
#include <cassert>
#include <iostream>

class THaTrack;
class THaBenchmark;
class TClonesArray;
class THashTable;

#ifdef EVENT_DISPLAY
#include <TLine.h>
#include <TPaveText.h>
#include <TCanvas.h>
#include <TList.h>
#include <TGaxis.h>
#include <TPaletteAxis.h>
#include <TH2F.h>
class TGraphErrors;
class TLegend;
class TStyle;
class TH1F;
#endif

using std::vector;
using namespace std;

namespace TreeSearch {
  class ReadoutPlane;
  class Projection;
  class Road;
  class ThreadCtrl;  // Defined in implementation

  typedef std::vector<Road*> Rvec_t;
  typedef std::set<Road*>    Rset_t;

  extern const Double_t kBig;

  class GEM : public THaTrackingDetector {
  public:
    GEM( const char* name, const char* description = "", 
	  THaApparatus* app = 0 );
    virtual ~GEM();

    virtual void    Clear( Option_t* opt="" );
    virtual Int_t   Decode( const THaEvData& );
    virtual EStatus Init( const TDatime& date );
    virtual Int_t   CoarseTrack( TClonesArray& tracks );
    virtual Int_t   FineTrack( TClonesArray& tracks );
    virtual Int_t   DefineVariables( EMode mode = kDefine );
    virtual void    Print(const Option_t* opt) const;
    virtual void    SetDebug( Int_t level );

    void            EnableEventDisplay( Bool_t enable = true );
    const pdbl_t&   GetChisqLimits( UInt_t i ) const;

#ifdef TESTCODE
    virtual Int_t   Begin( THaRunBase* r=0 );
    virtual Int_t   End( THaRunBase* r=0 );

    Int_t           GetEvNum() const { return fEvNum; }
    vector<TreeSearch::ReadoutPlane*>& GetListOfPlanes() { return fPlanes; }
    vector<TreeSearch::Projection*>& GetListOfProjections() { return fProj; }
#endif
    // Analysis control flags. Set via database.
    enum {
      kDoADCCut   = BIT(14), // Use ADC cut in decoder (defined in planes)
      k3dCorrAmpl = BIT(17), // Use amplitude correation 3D matching algorithm
      k3dFastMatch= BIT(18), // Use fast 3D matching algorithm (auto detected)
      kEventDisplay=BIT(19), // Support event display
      kDoCoarse   = BIT(20), // Do coarse tracking (if disabled, decode only)
      kDoFine     = BIT(21)  // Do fine tracking (implies kDoCoarse)
    };
#ifdef EVENT_DISPLAY
    class MC_Track {
    public:
    MC_Track(Double_t trackx , Double_t tracky , Double_t trackth , Double_t trackph):tr_x(trackx),tr_y(tracky),tr_th(trackth),tr_ph(trackph){};
      Double_t tr_x;
      Double_t tr_y;
      Double_t tr_th;
      Double_t tr_ph;
      Int_t GetStripNum(Double_t StripStart , Double_t StripPitch , Double_t PlaneZ , EProjType type ) { 
	Double_t PlaneHitPos = 0.0;
	if(string(kProjParam[type].name) == "x") {
	  PlaneHitPos = tr_x + tr_th * PlaneZ;
	} else {
	  PlaneHitPos = tr_y + tr_ph * PlaneZ;	  
	}
	Int_t HitStrip = Int_t((PlaneHitPos - StripStart)/StripPitch);
	return HitStrip;
      }
    };
    MC_Track *MCTrackInfo;
    MC_Track* GetMCTrack() const { return MCTrackInfo;}
    void  UpdateCanvas() {
      ccDisplay->cd(1);
      DrawXlines();
      ccDisplay->cd(2);
      DrawYlines();
      ccDisplay->Update();   
    }
#endif
  protected:
    friend class ReadoutPlane;
    class TrackFitWeight;
    struct FitRes_t {
      vector<Double_t> coef;
      Double_t matchval;
      Double_t chi2;
      Int_t    ndof;
      Rvec_t*  roads;
      FitRes_t() : roads(0) {}
    };

    typedef std::vector<ReadoutPlane*>  Rpvec_t;
    typedef std::vector<Projection*> Prvec_t;

    Rpvec_t        fPlanes;      // Readout planes
    Prvec_t        fProj;        // Plane projections
    Rpvec_t        fCalibPlanes; // Planes in calibration mode

    THashTable*    fCrateMap;    // Map of GEM DAQ modules

    // Multithread support
    Int_t          fMaxThreads;  // Maximum simultaneously active threads
    ThreadCtrl*    fThreads;     //! Thread controller

    // Paremeters for 3D projection matching
    Double_t       f3dMatchvalScalefact; // Correction for fast 3D matchval
    Double_t       f3dMatchCut;          // Maximum allowed 3D match error
    UInt_t         fMaxCorrMismatches;   // Max # planes w/o amplitude match
    Double_t       fMaxCorrNsigma;       // Amplitude correlation cutoff (#sig)

    // Track fit cut parameters
    Int_t          fMinNdof;     // Minimum number of points in fit - 4
    vec_pdbl_t     fChisqLimits; // lo/hi confidence interval limits on Chi2

    // Event data
    Int_t          fFailNhits; // Too many hits in wire plane(s)
    Int_t          fFailNpat;  // Too many treesearch patterns found
#ifdef TESTCODE
    UInt_t         fNcombos;   // Number of road combinations tried
    UInt_t         fN3dFits;   // Number of track fits done (=good road combos)
    Int_t          fEvNum;     // Current event number
    Double_t       t_track, t_3dmatch, t_3dfit, t_coarse; // times in us
#endif

    void      Add3dMatch( const Rvec_t& selected, Double_t matchval,
			  std::list<std::pair<Double_t,Rvec_t> >& combos_found,
			  Rset_t& unique_found ) const;
    void      FindNearestHits( ReadoutPlane* wp, const THaTrack* track,
			       const Rvec_t& roads ) const;
    void      FitErrPrint( Int_t err ) const;
    Int_t     FitTrack( const Rvec_t& roads, vector<Double_t>& coef,
			Double_t& chi2, TMatrixDSym* coef_covar = 0 ) const;
    template< typename Action > static
    Action    ForAllTrackPoints( const Rvec_t& roads, 
				 const vector<Double_t>& coef, Action action );
    UInt_t    MatchRoads( vector<Rvec_t>& roads,
			  std::list<std::pair<Double_t,Rvec_t> >& combos_found,
			  Rset_t& unique_found );
			  
    THaTrack* NewTrack( TClonesArray& tracks, const FitRes_t& fit_par );
    Bool_t    PassTrackCuts( const FitRes_t& fit_par ) const;

    // Helper functions for getting DAQ module parameters - used by Init
    UInt_t    LoadDAQmodel( THaDetMap::Module* m ) const;
    UInt_t    GetDAQnchan( THaDetMap::Module* m ) const;

    // Podd interface
    virtual Int_t   ReadDatabase( const TDatime& date );

#ifdef EVENT_DISPLAY
    TCanvas *ccDisplay;
    TH2F    *hHitsX;
    TH2F    *hHitsY;
    TLegend *legend;
    TCanvas *ccShowADC;
    vector<TLine*> vXlines;
    vector<TLine*> vYlines;
    TGaxis *axis1;
    TGaxis *axis2;
    TPaletteAxis *palette1;
    TPaletteAxis *palette2;

    void     DrawXlines() {
      for(vector<TLine*>::iterator iter = vXlines.begin(); iter!=vXlines.end(); iter++) (*iter)->Draw();
      ccDisplay->GetPad(1)->Update();
      TPaveText *title = static_cast<TPaveText*>(ccDisplay->GetPad(1)->GetListOfPrimitives()->FindObject("title"));
      title->SetX1NDC(0.1);
      title->SetX2NDC(0.8);
      title->SetY1NDC(0.93);
      title->SetY2NDC(1.0);
      axis1->Draw();
      palette1 = static_cast<TPaletteAxis*>(hHitsX->GetListOfFunctions()->FindObject("palette"));
      if(palette1) {
	palette1->SetX1NDC(0.88);
	palette1->SetX2NDC(0.92);
	palette1->SetY1NDC(0.1);
	palette1->SetY2NDC(0.9);
	palette1->Draw();
      }
    }
    void     DrawYlines() {
      for(vector<TLine*>::iterator iter = vYlines.begin(); iter!=vYlines.end(); iter++) (*iter)->Draw();
      ccDisplay->GetPad(2)->Update();
      axis2->Draw();
      TPaveText *title = static_cast<TPaveText*>(ccDisplay->GetPad(2)->GetListOfPrimitives()->FindObject("title"));
      if(title) {
	title->SetX1NDC(0.1);
	title->SetX2NDC(0.35);
	title->SetY1NDC(0.93);
	title->SetY2NDC(1.0);
      }
      palette2 = static_cast<TPaletteAxis*>(hHitsY->GetListOfFunctions()->FindObject("palette"));
      if(palette2) { 
	palette2->SetX1NDC(0.88);
	palette2->SetX2NDC(0.92);
	palette2->SetY1NDC(0.1);
	palette2->SetY2NDC(0.9);
	palette2->Draw();
      }
    }
#endif
    ClassDef(GEM,0)   // Tree search track reconstruction for GEM trackers
  };
  using namespace std;  
  //___________________________________________________________________________
  inline const pdbl_t& GEM::GetChisqLimits( UInt_t i ) const
  { 
    assert( i < fChisqLimits.size() );
    //    cout << " Test i ->> " << i << " assert = " << fChisqLimits.size() <<endl;
    return fChisqLimits[i];
  }

}

///////////////////////////////////////////////////////////////////////////////

#endif
