#ifndef ROOT_TreeSearch_Helper
#define ROOT_TreeSearch_Helper

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// TreeSearch::Helper                                                        //
//                                                                           //
// Helper classes and functions                                              //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include "Rtypes.h"
#include <vector>
#include <cassert>
#include <functional>
#include <algorithm>
#include <stdexcept>

namespace TreeSearch {

  //___________________________________________________________________________
  template< typename VectorElem > inline void 
  NthCombination( UInt_t n, const std::vector<std::vector<VectorElem> >& vec,
		  std::vector<VectorElem>& selected )
  {
    // Get the n-th combination of the elements in "vec" and
    // put result in "selected". selected[k] is one of the
    // vec[k].size() elements in the k-th plane.
    //
    // This function is equivalent to vec.size() nested/recursive loops
    // over the elements of each of the vectors in vec.

    using std::vector;

    assert( !vec.empty() );

    UInt_t vsiz = vec.size(), k;
    selected.resize( vsiz );
    typename vector< vector<VectorElem> >::const_iterator iv = vec.begin();
    typename vector<VectorElem>::iterator is = selected.begin();

    while( iv != vec.end() ) {
      typename std::vector<VectorElem>::size_type npt = (*iv).size();
      assert(npt);
      if( npt == 1 )
	k = 0;
      else {
	k = n % npt;
	n /= npt;
      }
      // Copy the selected element
      (*is) = (*iv)[k];
      ++iv;
      ++is;
    }
  }

  //___________________________________________________________________________
  template< typename Container > class SizeMul :
    public std::binary_function< typename Container::size_type, Container,
				 typename Container::size_type >
  {
    // Function object to get the product of the sizes of the containers in a 
    // container, ignoring empty ones
  private:
    typedef typename Container::size_type csiz_t;
    static size_t highestBit( register csiz_t a )
    {
      // Return highest bit set of argument.
      // A lookup table would break generality needed for templating
      register size_t bits = 0;
      if( !(a > 0) ) a = -a;
      while( a ) { ++bits; a >>= 1; }
      return bits;
    }
    size_t arg_size;  // Number of bits of csiz_t type (excluding sign)
  public:
    SizeMul() : arg_size(8*sizeof(csiz_t))
    {
      csiz_t n = -1;
      if( !( n > 0 ) ) --arg_size;  // signed type
    }
    csiz_t operator() ( csiz_t val, const Container& c ) const
    { 
      if( c.empty() ) return val;
      csiz_t arg = c.size();
      if( arg != 1 && val != 0 && val != 1 &&
	  highestBit(val)+highestBit(arg) > arg_size ) // conservative check
	throw std::overflow_error("result too large");
      return val * arg;
    }
  };

  //___________________________________________________________________________
  // Iterator over unique combinations of k out of N elements. The element 
  // numbers are accessible in a vector<int> of size k.
  class UniqueCombo {
    typedef std::vector<int> vint_t;
  public:
    UniqueCombo( int N, int k ) : fN(N), fGood(true)
    { assert( 0<=k && k<=N ); for( int i=0; i<k; ++i ) fCurrent.push_back(i); }
    // Default copy and assignment are fine

    UniqueCombo& operator++()
    {
      if( fCurrent.empty() )
	fGood = false;
      else
	fGood = recursive_plus( fCurrent.size()-1 );
      return *this;
    }
    const UniqueCombo operator++(int)
    { UniqueCombo clone(*this); ++*this; return clone; }
    const vint_t& operator()() const { return fCurrent; }
    const vint_t& operator*()  const { return fCurrent; }
    bool operator==( const UniqueCombo& rhs ) const
    { return ( fN == rhs.fN and fCurrent == rhs.fCurrent ); }
    bool operator!=( const UniqueCombo& rhs ) const { return !(*this==rhs); }
    operator bool() const  { return fGood; }
    bool operator!() const { return !((bool)*this); }
    
  private:
    bool recursive_plus( vint_t::size_type pos ) {
      if( fCurrent[pos] < fN+int(pos-fCurrent.size()) ) {
	++fCurrent[pos];
	return true;
      }
      if( pos==0 )
	return false;
      if( recursive_plus(pos-1) ) {
	fCurrent[pos] = fCurrent[pos-1]+1;
	return true;
      }
      return false;
    }
    int    fN;
    vint_t fCurrent;
    bool   fGood;
  };

  //___________________________________________________________________________
  struct DeleteObject {
    template< typename T >
    void operator() ( const T* ptr ) const { delete ptr; }
  };

  //___________________________________________________________________________
  template< typename Container >
  inline void DeleteContainer( Container& c )
  {
    // Delete all elements of given container of pointers
    for_each( c.begin(), c.end(), DeleteObject() );
    c.clear();
  }

  //___________________________________________________________________________
  template< typename ContainerOfContainers >
  inline void DeleteContainerOfContainers( ContainerOfContainers& cc )
  {
    // Delete all elements of given container of containers of pointers
    for_each( cc.begin(), cc.end(), 
	      DeleteContainer<typename ContainerOfContainers::value_type> );
    cc.clear();
  }
  
///////////////////////////////////////////////////////////////////////////////

} // end namespace TreeSearch


#endif
