//*-- Author:  Ole Hansen<mailto:ole@jlab.org>; Jefferson Lab; 11-Jan-2010
//
#ifndef ROOT_TreeSearch_ReadoutPlane
#define ROOT_TreeSearch_ReadoutPlane

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// TreeSearch::ReadoutPlane                                                  //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include "Types.h"
#include "THaSubDetector.h"
#include "TClonesArray.h"
#include "TVector2.h"
#include "TMath.h"
#include "TBits.h"
#include <vector>
#include <string>
#include <map>
#include <cassert>
#include <functional>

using std::vector;
using std::string;

class TVector2;
class TH1;

namespace TreeSearch {

  class TimeToDistConv;
  class Projection;
  class GEM;
  class Hit;
  class FitCoord;
  extern const Double_t kBig;

  class ReadoutPlane : public THaSubDetector {

    typedef vector<Float_t> Vflt_t;

  public:
    ReadoutPlane( const char* name, const char* description = "", 
	       THaDetectorBase* parent = 0 );
    ReadoutPlane() : fPartner(0), fProjection(0), fGEM(0), fADC(0), fADCped(0),
		     fHits(0), fFitCoords(0) {} // For ROOT RTTI
    virtual ~ReadoutPlane();

    virtual void    Clear( Option_t* opt="" );
    virtual Int_t   Decode( const THaEvData& );
    virtual void    Print( Option_t* opt="" ) const;

#ifdef TESTCODE
    virtual Int_t   Begin( THaRunBase* r=0 );
    virtual Int_t   End( THaRunBase* r=0 );
#endif

//     virtual Int_t   Compare ( const TObject* obj ) const;
//     virtual Bool_t  IsSortable () const { return kTRUE; }

    FitCoord*       AddFitCoord( const FitCoord& coord );
    Bool_t          Contains( Double_t x, Double_t y ) const;
    Bool_t          Contains( const TVector2& point ) const;
    void            EnableCalibration( Bool_t enable = true );
    Double_t        GetAmplSigma( Double_t ampl ) const;
    EProjType       GetType()        const { return fType; }
    Double_t        GetZ()           const { return fOrigin.Z(); }
    Projection*     GetProjection()  const { return fProjection; }
    ReadoutPlane*   GetPartner()     const { return fPartner; }
    void            SetPartner( ReadoutPlane* p );
    const char*     GetPartnerName() const { return fPartnerName.Data(); }
    GEM*            GetGEM()         const { return fGEM; }
    Double_t        GetResolution()  const { return fResolution; }
    Double_t        GetMaxSlope()    const; 
    Double_t        GetStripStart()  const { return fStripStart+fCoordOffset; }
    Double_t        GetStripPitch()  const { return fStripPitch; }
#ifdef EVENT_DISPLAY
    Double_t        GetStripCenter( Int_t istrip ) const { return fStripStart+fCoordOffset+fStripPitch*istrip; }
#endif
    TSeqCollection* GetHits()        const { return fHits; }
    Int_t           GetNhits()       const { return fHits->GetLast()+1; }
    TSeqCollection* GetCoords()      const { return fFitCoords; }
    Int_t           GetNcoords()     const { return fFitCoords->GetLast()+1; }
    UInt_t          GetPlaneNum()    const { return fPlaneNum; }
    UInt_t          GetDefinedNum()  const { return fDefinedNum; }
    Bool_t          IsCalibrating()  const { return TestBit(kCalibrating); }
    Bool_t          IsRequired()     const;
    Double_t        GetRawOcc()      const { return fRawOcc; }
    Double_t        GetOccupancy()   const { return fOccupancy; }

    void            SetPlaneNum( UInt_t n )   { fPlaneNum = n; }
    void            SetDefinedNum( UInt_t n ) { fDefinedNum = n; }
    void            SetProjection( Projection* p ) 
    { fProjection = p; UpdateOffset(); }
    void            SetRequired( Bool_t enable = true );
    void            UpdateOffset();

    vector<vector<Float_t> >  GetADCSamples() const { return fADCsamp; }
    // Helper functors for STL algorithms...
    struct ZIsLess
      : public std::binary_function< ReadoutPlane*, ReadoutPlane*, bool >
    {
      bool operator() ( const ReadoutPlane* a, const ReadoutPlane* b ) const
      { 
	assert( a && b );
	assert( a->GetDefinedNum() != b->GetDefinedNum() );
	if( a->GetZ() < b->GetZ() ) return true;
	if( b->GetZ() < a->GetZ() ) return false;
	return( a->GetDefinedNum() < b->GetDefinedNum() );
      }
    };

    class NameEquals : public std::unary_function< ReadoutPlane*, bool > {
    public:
      NameEquals( const char* s ) : name(s?s:"") {}
      bool operator() ( const ReadoutPlane* rp ) const
      { assert(rp); return ( name == rp->GetName() ); }
    private:
      string name;
    };
#ifdef EVENT_DISPLAY
    std::map<Int_t,Float_t*>  mStripADC;
    std::map<Int_t,Float_t*>  GetStripADCMap() const { return mStripADC;}
#endif

  protected:

    // Geometry, configuration
    UInt_t        fPlaneNum;    // Ordinal of this plane within its projection
    UInt_t        fDefinedNum;  // Ordinal of this plane in planeconfig string
    EProjType     fType;        // Plane type (x,y,u,v)
    Double_t      fStripStart;  // Position of 1st strip (m)
    Double_t      fStripPitch;  // Strip pitch (assumed constant) (m)
    Double_t      fCoordOffset; // Strip coord offset wrt GEM due to fOrigin
    ReadoutPlane* fPartner;     //! Partner plane (if 2-D readout)
    Projection*   fProjection;  //! The projection that we belong to
    GEM*          fGEM;         //! Our parent detector
    TString       fPartnerName; // Name of partner plane in same 2-D readout

    // Hardware channel mapping
    enum EChanMapType { kOneToOne, kReverse, kGassiplexAdapter1,
			kGassiplexAdapter2, kTable };

    EChanMapType  fMapType;     // Type of hardware channel mapping to use
    vector<Int_t> fChanMap;     // [fNelem] Optional hardware channel mapping
    Int_t         MapChannel( Int_t idx ) const;
    // Parameters, calibration, flags
    Double_t      fResolution;  // Drift distance resolution (sigma) (m)
    UInt_t        fMaxClusterSize;// Maximum size of a clean cluster of strips
    Double_t      fMinAmpl;     // ADC threshold for strips to be active
    Double_t      fSplitFrac;   // Percentage of amplitude swing necessary
                                // for detecting cluster maximum/minimum
    UInt_t        fMaxSamp;     // Maximum # ADC samples per channel
    UInt_t        fMaxHits;     // Maximum # hits before flagging decode error

    Vflt_t        fPed;         // [fNelem] Per-channel pedestal values
    TBits         fBadChan;     // Bad channel map
    Double_t      fAmplSigma;   // Sigma of hit ampliude distribution

    // Event data, hits etc.
    vector<Vflt_t> fADCsamp;    // [fNelem] Raw ADC values/time samples
    // TODO: convert to vectors once global variable system can handle them
    Float_t*      fADC;         // [fNelem] Raw ADC integrals
    Float_t*      fADCped;      // [fNelem] Pedestal/noise-subtracted ADC vals
    TClonesArray* fHits;        // Cluster data (groups of hits)
    TClonesArray* fFitCoords;   // Cluster coordinates used by good road fits
    Double_t      fDnoise;      // Event-by-event noise (avg below fMinAmpl)

    UInt_t        fNhitStrips;  // Statistics: strips with any data
    UInt_t        fNsigStrips;  // Statistics: strips > 0
    UInt_t        fNRawHits;    // Statistics: strips > fMinAmpl
    Double_t      fRawOcc;      // Statistics: raw occupancy fNsigStrips/fNelem
    Double_t      fOccupancy;   // Statistics: occupancy fNRawHits/fNelem

#ifdef TESTCODE
    // Optional diagnostics
    TH1*          fHitMap;      // Histogram of active strip numbers
    TH1*          fADCMap;      // Histogram of strip numbers weighted by ADC
#endif
    
    // Podd interface
    virtual Int_t ReadDatabase( const TDatime& date );
    virtual Int_t DefineVariables( EMode mode = kDefine );

    // Bits for ReadoutPlanes
    enum {
      kIsRequired  = BIT(14), // Tracks must have a hit in this plane
      kCalibrating = BIT(15), // Plane in calibration mode (implies !required)
      kDoNoise     = BIT(16)  // Correct data for common-mode noise
#ifdef TESTCODE
     ,kDoHistos    = BIT(23)  // Generate hitmap and adcmap histograms
#endif
    };

    ClassDef(ReadoutPlane,0)  // One GEM readout plane coordinate direction
  };

  //___________________________________________________________________________
  inline
  Bool_t ReadoutPlane::Contains( Double_t x, Double_t y ) const
  {
    // Check if the given point is within the active area of this wire plane.
    // Coordinates are relative to the GEM origin. Time-critical, may be
    // be called O(1e5) per event

    //TODO: allow for (small) rotation due to misalignment?

    return ( TMath::Abs( x-fOrigin.X() ) < fSize[0] and
	     TMath::Abs( y-fOrigin.Y() ) < fSize[1] );
  }

  //___________________________________________________________________________
  inline
  Bool_t ReadoutPlane::Contains( const TVector2& point ) const
  {
    // Same as Contains(x,y), but using a TVector2 as input

    return Contains( point.X(), point.Y() );
  }

  //___________________________________________________________________________
  inline
  void ReadoutPlane::EnableCalibration( Bool_t enable )
  {
    // Enable/disable calibration mode flag

    SetBit( kCalibrating, enable );
  }
  
  //___________________________________________________________________________
  inline
  Bool_t ReadoutPlane::IsRequired() const
  { 
    return ( not IsCalibrating() and TestBit(kIsRequired) );
  }

  //___________________________________________________________________________
  inline
  void ReadoutPlane::SetRequired( Bool_t enable )
  {
    // Enable/disable calibration mode flag

    SetBit( kIsRequired, enable );
  }
  
///////////////////////////////////////////////////////////////////////////////

} // end namespace TreeSearch


#endif
