//*-- Author :    Jens-Ole Hansen (ole@jlab.org)    23-Jul-2010

/////////////////////////////////////////////////////////////////////
//
//   SBSsim
//
//   Common class definitions for SBSsim package
//
/////////////////////////////////////////////////////////////////////

#include "SBSsim.h"
#include "TSystem.h"
#include "TMath.h"
#include "TClonesArray.h"
#include "TString.h"

#include <iostream>
#include <cassert>

using namespace std;

//-----------------------------------------------------------------------------
void SBSsimTrack::Print( const Option_t* opt ) const
{
  //TObject::Print(opt);
  cout << "track: type = " << fType << ", hit1 = " << fHit1 
       << ", PID = " << fPID << ", t0 = " << fTimeOffset << endl;
  cout << "  Origin    = (" << fOrigin.X() << ", " << fOrigin.Y() << ", " 
       << fOrigin.Z() << ")" << endl;
  cout << "  Momentum  = (" << fMomentum.Px() << ", " 
       << fMomentum.Py() << ", " << fMomentum.Pz() << ")" << endl;
}

//-----------------------------------------------------------------------------
SBSsimEvent::SBSsimEvent()
  : fEvtNum(0), fNhit(0), fNsignal(0), fTracks(0), fNchan(0)
{
}			     

//-----------------------------------------------------------------------------
SBSsimEvent::~SBSsimEvent()
{
  delete fTracks;
}

//-----------------------------------------------------------------------------
void SBSsimEvent::Clear( const Option_t* /*opt*/ )
{
  // Clear the event in preparation for reading next tree entry

  fEvtNum = 0;
  fNhit = fNsignal = 0;
  fTracks->Clear();
  fNchan = 0;
}

//-----------------------------------------------------------------------------
void SBSsimEvent::Print( const Option_t* opt ) const
{
  // Print current event info

  cout << ">>>>> =====================================" << endl;
  cout << "Event number:          " << fEvtNum << endl;
  cout << "Number of GEM hits:    " << fNchan  << endl;

  TString sopt(opt);
  bool do_hit    = sopt.Contains("hit", TString::kIgnoreCase);
  bool do_signal = sopt.Contains("signal", TString::kIgnoreCase);
  bool do_backgr = sopt.Contains("backgr", TString::kIgnoreCase);
  bool do_pileup = sopt.Contains("pileup", TString::kIgnoreCase);
  if( do_hit || do_signal || do_pileup ) {
    for( Int_t i=0; i<fNchan; ++i ) {
      if( do_hit ||
	  (do_signal && fType[i] == 1) || 
	  (do_backgr && fType[i] == 1) || 
	  (do_pileup && fType[i] == 3) ) {
	cout << "hit = " << i
	     << ", plane = " << fPlane[i]
	     << ", axis = "  << fAxis[i]
	     << ", strip = " << fStrip[i]
	     << ", type = "  << fType[i]
	     << ", adc = ";
	for( int k=0; k<MC_MAXSAMP; k++ ) {
	  cout << fADC[k][i];
	  if( k+1 != MC_MAXSAMP ) cout << ", ";
	}
	cout << endl;
      }
    }
  }

  if( sopt.Contains("track", TString::kIgnoreCase) && fTracks ) {
    for( Int_t i=0; i<fTracks->GetLast()+1; ++i ) {
      fTracks->UncheckedAt(i)->Print(opt);
    }
  }
}

//-----------------------------------------------------------------------------
SBSsimParameters::SBSsimParameters()
// FIXME: update for SBSsim
 //set defaults conditions
    : filename("vdctracks.root"), numTrials(10000), 
      numWires(368), noiseSigma(4.5), noiseMean(0.0), wireEff(1.0), tdcTimeLimit(900.0), 
      emissionRate(0.000002), probWireNoise(0.0), x1(-1.8), x2(-0.1),ymean(0.0), 
      ysigma(0.01), z0(-1.0), 
      pthetamean(TMath::Pi()/4.0), pthetasigma(atan(1.1268) - TMath::Pi()/4.0), 
      pphimean(0.0), pphisigma(atan(0.01846)), pmag0(1.0), tdcConvertFactor(2.0), 
      cellWidth(0.0042426), cellHeight(0.026)
{
  // Constructor - determine the name of the database file

  const char* dbDir       = gSystem->Getenv("DB_DIR");
  const char* analyzerDir = gSystem->Getenv("ANALYZER");
  if( dbDir ) {
    databaseFile = dbDir;
    databaseFile += "/";
  }
  else if( analyzerDir ) {
    databaseFile = analyzerDir;
    databaseFile += "/DB/";
  }
  else
    databaseFile = "/u/home/orsborn/vdcsim/DB/";
  databaseFile += "20030415/db_L.vdc.dat";
}

//-----------------------------------------------------------------------------
SBSsimParameters::~SBSsimParameters()
{
  // Destructor
}

//-----------------------------------------------------------------------------
void SBSsimParameters::Print( const Option_t* opt ) const
{
  
}

//-----------------------------------------------------------------------------
Int_t SBSsimParameters::ReadDatabase(Double_t* timeOffsets)
{
  //open the file as read only
  FILE* file = fopen(databaseFile, "r");
  if( !file ) return 1;


  // Use default values until ready to read from database

  const int LEN = 100;
  char buff[LEN];
  
  for( int j = 0; j<4; j++ ) {
    rewind(file);
    // Build the search tag and find it in the file. Search tags
    // are of form [ <prefix> ], e.g. [ R.vdc.u1 ].
    TString tag(Prefixes[j]); tag.Prepend("["); tag.Append("]"); 
    TString line, tag2(tag);
    tag.ToLower();
    // cout << "ReadDatabase:  tag = " << tag.Data() << endl;
    bool found = false;
    while (!found && fgets (buff, LEN, file) != NULL) {
      char* buf = ::Compress(buff);  //strip blanks
      line = buf;
      delete [] buf;
      if( line.EndsWith("\n") ) line.Chop();  //delete trailing newline
      line.ToLower();
      if ( tag == line ) 
	found = true;
    }
    if( !found ) {
      fclose(file);
      return 1;
    }
    
    for(int i = 0; i<5; i++)
      fgets(buff, LEN, file);  //skip 5 lines to get to drift velocity

    //read in drift velocity
    Double_t driftVel = 0.0;
    fscanf(file, "%lf", &driftVel);
    driftVelocities[j] = driftVel/1e9;   //convert to m/ns

    fgets(buff, LEN, file); // Read to end of line
    fgets(buff, LEN, file); // Skip line
  
    //Found the offset entries for this plane, so read time offsets and wire numbers

    for (int i = j*numWires; i < (j+1)*numWires; i++) {
      int wnum = 0;
      Double_t offset = 0.0;
      fscanf(file, " %d %lf", &wnum, &offset);
      timeOffsets[i] = offset;
    }
  }

  fclose(file);
  return 0;
}

//-----------------------------------------------------------------------------
ClassImp(SBSsimEvent)
ClassImp(SBSsimParameters)
ClassImp(SBSsimTrack)
