/////////////////////////////////////////////////////////////////////
//
//   SBSsim
//
//   Common class definitions for SBSsim package
//
//   Jens-Ole Hansen (ole@jlab.org)  July 2010
//
/////////////////////////////////////////////////////////////////////

#ifndef SBSsim_
#define SBSsim_

#include "TVector3.h"
#include "TString.h"

class TClonesArray;

//-----------------------------------------------------------------------------
class SBSsimParameters : public TObject {
public:
  SBSsimParameters();
  virtual ~SBSsimParameters();

  virtual void Print(Option_t* opt="") const;

  TString filename;            //name of output file
  TString textFile;            //name of output text file (holds track information)
  int numTrials;               //number of events generated in monte carlo
  Double_t wireHeight[4];      // Height of each wire in Z coord
  Double_t driftVelocities[4]; // Drift velocity in Z coord
  TString Prefixes[4];         //array of prefixes to use in reading database
  TString databaseFile;        //name of database file to read
  int numWires;                //number of wires in each chamber
  Double_t noiseSigma;         //sigma value of noise distribution
  Double_t noiseMean;          //mean of noise distribution 
  Double_t wireEff;            //wire efficiency (decimal form)
  Double_t tdcTimeLimit;       //window of time the tdc reads (in ns)
  Double_t emissionRate;       //rate particles are emitted from target (in particles/ns)
  Double_t probWireNoise;      //probability of random chamber wires firing
  Double_t x1;      
  Double_t x2;           //x coordinate limits for origin of tracks
  Double_t ymean;
  Double_t ysigma;       //mean and sigma value for origin  y coord. distribution
  Double_t z0;           //z coordinate of track origin
  Double_t pthetamean;
  Double_t pthetasigma;  //mean and sigma for momentum theta distribution 
  Double_t pphimean;
  Double_t pphisigma;    //mean and sigma of momentum phi distribution
  Double_t pmag0;        //magnitude of momentum
  Double_t tdcConvertFactor;   //width of single time bin (in ns), used to convert from time to tdc values
  Double_t cellWidth;    //width of cell (i.e. horizontal spacing between sense wires)
  Double_t cellHeight;   //height of cell (i.e. vertical spacing between u and v wire planes)

  Int_t ReadDatabase(Double_t* timeOffsets);
  
  ClassDef(SBSsimParameters, 1) // simulation parameters
};

//-----------------------------------------------------------------------------
class SBSsimTrack : public TObject {
 public: 
  SBSsimTrack() : fType(0), fLayer(0), fPID(0), fTimeOffset(0.0), fHit1(-1) {}
  SBSsimTrack( Int_t type, Int_t layer, Int_t pid, Double_t toff )
    : fType(type), fLayer(layer), fPID(pid), fTimeOffset(toff), fHit1(-1) {}

  Int_t    fType;        // Track type  (0=trigger, 1=accidental/bckgr,
                         //   2=delta ray, 3=other)
  Int_t    fLayer;       // Material type where track originated (unused)
  Int_t    fPID;         // Track particle ID (PDG)
  Double_t fTimeOffset;  // Time offset of the track (s)
  TVector3 fOrigin;      // Origin (m)
  TVector3 fMomentum;    // Momentum (GeV)
  Int_t    fHit1;        // Chamber where this track makes its first hit

  Double_t TX()     { return fOrigin.X(); }
  Double_t TY()     { return fOrigin.Y(); }
  Double_t TTheta() { return fMomentum.Px()/fMomentum.Pz(); }
  Double_t TPhi()   { return fMomentum.Py()/fMomentum.Pz(); }
  Double_t P()      { return fMomentum.Mag(); }

  virtual void Print( const Option_t* opt="" ) const;
  
  ClassDef(SBSsimTrack, 1) // Simluated SBS track
};

//-----------------------------------------------------------------------------
// class SBSsimWireHit : public TObject {
// public:
//   SBSsimWireHit()
//     : wirenum(0), type(0), rawTime(0), rawTDCtime(0), time(0), distance(0), wireFail(false) {}

//   virtual void Print( const Option_t* opt="" ) const;


//   Int_t wirenum;      // Wire number
//   Int_t type;         // flag for tyep of hit. actual = 0, noise = 1
//   Double_t rawTime;    // Time of wire hit in nanoseconds
//   Int_t rawTDCtime; //tdc time w/out noise
//   Int_t time;       //tdctime w/additional noise
//   Double_t distance;   //drift distance 
//   bool wireFail;      //set to true when the wire fails
//   Double_t pos;       //wire position (m) along u(v) coordinate

//   ClassDef(SBSsimWireHit, 1) // Simple Wirehit class
// };

//-----------------------------------------------------------------------------
// Kludgy hardcoded parameters necessary because I can't get ROOT to allocate
// arrays dynamically via TTree::GetEntry
#define MC_MAXH  10000
#define MC_MAXC  20000
#define MC_MAXSAMP 10

class SBSsimEvent : public TObject {
public:
  SBSsimEvent();
  virtual ~SBSsimEvent();

  // Bookkeeping 
  Int_t     fEvtNum;        // Event number

  // Truth data
  // Primary track data (in world coordinates)
  Float_t   fTrackX;
  Float_t   fTrackY;
  Float_t   fTrackZ;
  Float_t   fTrackMX;
  Float_t   fTrackMY;
  Float_t   fTrackMZ;

  // Note: fNsignal <= fNhits (equal if no background)
  Int_t     fNhit;             // Number of MC hits (sum over all planes)
  Int_t     fNsignal;          // Number of MC hits from trigger track (signal)
  // The first fNsignal elements in these arrays are from the signal
  Float_t   fHitX[MC_MAXH];    // Hit x-coordinate (mm)
  Float_t   fHitY[MC_MAXH];    // Hit y-coordinate (mm)
  Float_t   fHitZ[MC_MAXH];    // Hit z-coordinate (mm) TODO: garbage?
  Float_t   fHitMX[MC_MAXH];   // Hit x-momentum (MeV)
  Float_t   fHitMY[MC_MAXH];   // Hit y-momentum (MeV)
  Float_t   fHitMZ[MC_MAXH];   // Hit z-coordinate (MeV)
  Float_t   fCharge[MC_MAXH];  // Hit charge
  Float_t   fTime[MC_MAXH];    // Hit time
  Short_t   fChamber[MC_MAXH]; // Chamber number for hit
  Int_t     fPID[MC_MAXH];     // Hit partcile ID (PDG)

  TClonesArray*   fTracks;     //-> Tracks in this event

  // Digitized strip amplitude data
  Int_t     fNchan;            // Number of strips firing
  Short_t   fPlane[MC_MAXC];   // [fNchan] Plane number
  Short_t   fAxis[MC_MAXC];    // [fNchan] Strip direction (0 = x, 1 = y)
  Short_t   fStrip[MC_MAXC];   // [fNchan] Strip index
  Short_t   fType[MC_MAXC];    // [fNchan] Signal type (1=track, 2=bck, 3=both)
  Short_t   fTrack[MC_MAXC];   // [fNchan] Track index (-1=none)
  Short_t   fADC[MC_MAXSAMP][MC_MAXC]; // [fNchan] ADC samples

  virtual void Clear( const Option_t* opt="" );
  virtual void Print( const Option_t* opt="" ) const;

  ClassDef(SBSsimEvent, 2) // Simulated data for one event
};


#endif
