//*-- Author :    Jens-Ole Hansen (ole@jlab.org)    23-Jul-2010

/////////////////////////////////////////////////////////////////////
//
//   SBSsimDecoder
//   Decoder for SuperBigBite GEM simulation data
//
/////////////////////////////////////////////////////////////////////

#include "SBSsimDecoder.h"
#include "SBSsim.h"
#include "THaCrateMap.h"
#include "THaBenchmark.h"
#include "VarDef.h"
#include "TClonesArray.h"

using namespace std;

static const char* const MC_PREFIX = "MC.";

//-----------------------------------------------------------------------------
SBSsimDecoder::SBSsimDecoder() : fIsSetup(false)
{
  // Constructor

  DefineVariables();
}

//-----------------------------------------------------------------------------
SBSsimDecoder::~SBSsimDecoder() {

  DefineVariables( THaAnalysisObject::kDelete );
}

//-----------------------------------------------------------------------------
Int_t SBSsimDecoder::DefineVariables( THaAnalysisObject::EMode mode )
{
  const char* const here = "SBSsimDecoder::DefineVariables";

  if( mode == THaAnalysisObject::kDefine && fIsSetup ) 
    return THaAnalysisObject::kOK;
  fIsSetup = ( mode == THaAnalysisObject::kDefine );

  RVarDef vars[] = {
    { "tr.n",    "Number of tracks",     "GetNTracks()" },
    { "tr.x",    "Track origin x (m)",   "fTracks.SBSsimTrack.TX()" },
    { "tr.y",    "Track origin y (m)",   "fTracks.SBSsimTrack.TY()" },
    { "tr.th",   "Track tan(theta)",     "fTracks.SBSsimTrack.TTheta()" },
    { "tr.ph",   "Track tan(phi)",       "fTracks.SBSsimTrack.TPhi()" },
    { "tr.p",    "Track momentum (GeV)", "fTracks.SBSsimTrack.P() "},
    { "tr.hit1", "Chamber # of 1st hit", "fTracks.SBSsimTrack.fHit1" },
    { "tr.pid",  "Track PID (PDG)",      "fTracks.SBSsimTrack.fPID" },
    { "tr.t0",   "Track time offset",    "fTracks.SBSsimTrack.fTimeOffset" },
    { 0 }
  };
  
  Int_t ret = 
    THaAnalysisObject::DefineVarsFromList( vars, 
					   THaAnalysisObject::kRVarDef, 
					   mode, "", this, MC_PREFIX, here );

  return ret;
}

//-----------------------------------------------------------------------------
void SBSsimDecoder::Clear( Option_t* opt )
{
  // Clear track and plane data

  THaEvData::Clear();

  // Never delete the tracks, only clear the list. The tracks are deleted
  // in SBSsimRun::ReadEvent() by the call to event->Clear().
  fTracks.Clear("nodelete");
}

//-----------------------------------------------------------------------------
int SBSsimDecoder::LoadEvent(const int* evbuffer, THaCrateMap* map)
{
  // Fill crateslot structures with Monte Carlo event data in 'evbuffer'

  fMap = map;

  // Local copy of evbuffer pointer - any good use for it?
  buffer = evbuffer;

  // Cast the evbuffer pointer back to exactly the event type that is present
  // in the input file (in SBSsimRun). The pointer-to-integer is there
  // just for compatibility with the standard decoder.
  // Note: simEvent can't be constant - ROOT does not like to iterate
  // over const TList.
  SBSsimEvent* simEvent = (SBSsimEvent*)(evbuffer);

  if (first_decode) {
    init_cmap();     
    if (init_slotdata(map) == HED_ERR) return HED_ERR;
    first_decode = false;
  }
  if( fDoBench ) fBench->Begin("clearEvent");
  Clear();
  for( int i=0; i<fNSlotClear; i++ )
    crateslot[fSlotClear[i]]->clearEvent();
  if( fDoBench ) fBench->Stop("clearEvent");
  
  // FIXME: needed?
  evscaler = 0;
  event_length = 0;

  event_type = 1;
  event_num = simEvent->fEvtNum;
  recent_event = event_num;

  if( fDoBench ) fBench->Begin("physics_decode");

  // Decode the digitized data.  Populate crateslot array.
  for( Int_t i = 0; i < simEvent->fNchan; i++) {

    Int_t roc  = 2*simEvent->fPlane[i] + simEvent->fAxis[i];
    Int_t slot = simEvent->fStrip[i] / CHAN_PER_SLOT;
    Int_t chan = simEvent->fStrip[i] % CHAN_PER_SLOT;

    for( Int_t k = 0; k < MC_MAXSAMP; k++ ) {
      Int_t raw = simEvent->fADC[k][i];
      if (crateslot[idx(roc,slot)]->loadData("tdc",chan,raw,raw)
	  == SD_ERR) return HED_ERR;
    }
  }

  // Extract MC track info, so we can access it via global variables
  // The list of tracks is already part of the event - no need to generate
  // our own tracks here. 
  // FIXME: However, we have to copy the list because the global variable system 
  // cannot handle variable pointers.

  TClonesArray* tracks = simEvent->fTracks;
  for( Int_t i = 0; i < tracks->GetLast()+1; i++ ) {
    fTracks.Add( tracks->UncheckedAt(i) );
  }

  if( fDoBench ) fBench->Stop("physics_decode");

  // DEBUG:
  //cout << "SimDecoder: nTracks = " << GetNTracks() << endl;
  //fTracks.Print();

  return HED_OK;
}

//-----------------------------------------------------------------------------
ClassImp(SBSsimDecoder)
