//*-- Author :    Jens-Ole Hansen (ole@jlab.org)    23-Jul-2010

/////////////////////////////////////////////////////////////////////
//
//   SBSsimRun
//
//   Interface to a ROOT file with simulated SBS data
//
/////////////////////////////////////////////////////////////////////

#include "SBSsimRun.h"
#include "SBSsim.h"
#include <evio.h>

#include "TFile.h"
#include "TTree.h"
#include "TError.h"
#include "TClonesArray.h"
#include "TString.h"
#include "TMath.h"

#include <cassert>
#include <cstring>
#include <libgen.h>
#include <cstdlib>

using namespace std;

//-----------------------------------------------------------------------------
SBSsimRun::SBSsimRun(const char* filename, const char* description) :
  THaRunBase(description), rootFileName(filename), rootFile(0), tree(0), 
  event(0), nentries(0), entry(0)
{
  // Constructor

  // Default file name if none given
  if( rootFileName.IsNull() )
    rootFileName = "MCdata.root";

}

//-----------------------------------------------------------------------------
SBSsimRun::SBSsimRun(const SBSsimRun &run)
  : THaRunBase(run), nentries(0), entry(0)
{
  rootFileName = run.rootFileName;
  rootFile = 0;
  tree = 0;
  event = 0;
}

//-----------------------------------------------------------------------------
SBSsimRun& SBSsimRun::operator=(const THaRunBase& rhs)
{
  if (this != &rhs) {
    THaRunBase::operator=(rhs);
    if( rhs.InheritsFrom("SBSsimRun") )
      rootFileName = static_cast<const SBSsimRun&>(rhs).rootFileName;
    rootFile = 0;
    tree = 0;
    event = 0;
  }
  return *this;
}

//_____________________________________________________________________________
Int_t SBSsimRun::Compare( const TObject* obj ) const
{
  // Compare two SBSsimRuns. They are different if either their number
  // or their input file name differs.

  if (this == obj) return 0;
  const THaRunBase* rhs = dynamic_cast<const THaRunBase*>(obj);
  if( !rhs ) return -1;
  // operator< compares fNumber
  if( *this < *rhs )       return -1;
  else if( *rhs < *this )  return  1;
  const SBSsimRun* rhsr = dynamic_cast<const SBSsimRun*>(rhs);
  if( !rhsr ) return 0;
  if( rootFileName < rhsr->rootFileName ) return -1;
  else if( rhsr->rootFileName < rootFileName ) return 1;
  return 0;
}

//-----------------------------------------------------------------------------
Int_t SBSsimRun::Init()
{
  // Initialize the run. Sets run date, reads run database etc.

  // Set date to the day of the production, as far as relevant.
  // g1-g15: 01-Apr-10
  // g16:    08-Feb-11
  // g17:    02-Mar-11
  //
  // These dates are just a kludge to be able to read the correct section 
  // of the database.
  // NB: The "new" database system does not currently support "configurations"
  fAssumeDate = kTRUE;
  fDataSet |= kDate;
  // fDate is actually set in SBSsimRun::Open(), which is called through
  // this call to Init(). We can only detect the type of production when
  // we open the input file.
  Int_t ret = THaRunBase::Init();
  if( !ret ) {
    char* s = strdup(rootFileName);
    fName = basename(s);
    free(s);
    fNumber = 1;
  }
  return ret;
}

//-----------------------------------------------------------------------------
Int_t SBSsimRun::Open()
{
  rootFile = new TFile(rootFileName, "READ", "SBS MC data");
  if (!rootFile || rootFile->IsZombie()) {
    Close();
    return -1;
  }

  const char* const treeName = "jt_o";

  tree = static_cast<TTree*>(rootFile->Get(treeName));
  if( !tree ) {
    Error( "SBSsimRun::Open", "Tree %s does not exist in the input file.",
	   treeName );
    return -2;
  }
  tree->SetBranchStatus("*", kFALSE);

  // Create the event stucture by hand for now - the input file contains 
  // only the data members
  // This will happen automatically once the ROOT file contains event objects
  if( !event ) {
    event = new SBSsimEvent();

    event->fTracks = new TClonesArray("SBSsimTrack",1);

    tree->SetBranchStatus("EvtID", kTRUE);
    tree->SetBranchAddress("EvtID", &(event->fEvtNum) );

    tree->SetBranchStatus("digi.gem.nhit", kTRUE);
    tree->SetBranchAddress("digi.gem.nhit", &(event->fNhit) );
    tree->SetBranchStatus("digi.gem.nsignal", kTRUE);
    tree->SetBranchAddress("digi.gem.nsignal", &(event->fNsignal) );

    // The MC track truth information is only available as of production g17
    UInt_t found = 0;
    CLRBIT( fFlags, kHasTracks );
    tree->SetBranchStatus("digi.gem.par.x", kTRUE, &found);
    if( found > 0 ) {
      tree->SetBranchAddress("digi.gem.par.x", &(event->fTrackX) );
      tree->SetBranchStatus("digi.gem.par.y", kTRUE);
      tree->SetBranchAddress("digi.gem.par.y", &(event->fTrackY) );
      tree->SetBranchStatus("digi.gem.par.z", kTRUE);
      tree->SetBranchAddress("digi.gem.par.z", &(event->fTrackZ) );
      tree->SetBranchStatus("digi.gem.par.mx", kTRUE);
      tree->SetBranchAddress("digi.gem.par.mx", &(event->fTrackMX) );
      tree->SetBranchStatus("digi.gem.par.my", kTRUE);
      tree->SetBranchAddress("digi.gem.par.my", &(event->fTrackMY) );
      tree->SetBranchStatus("digi.gem.par.mz", kTRUE);
      tree->SetBranchAddress("digi.gem.par.mz", &(event->fTrackMZ) );
      SETBIT( fFlags, kHasTracks );
    } else {
      Warning( "SBSsimRun::Open", "No MC track data in input file %s",
	       rootFileName.Data() );
    }

    // Hit positions are only available as of production g16
    found = 0;
    CLRBIT( fFlags, kHasHitpos );
    tree->SetBranchStatus("digi.gem.hit.x", kTRUE, &found);
    if( found > 0 ) {
      tree->SetBranchAddress("digi.gem.hit.x", &(event->fHitX) );
      tree->SetBranchStatus("digi.gem.hit.y", kTRUE);
      tree->SetBranchAddress("digi.gem.hit.y", &(event->fHitY) );
      tree->SetBranchStatus("digi.gem.hit.z", kTRUE);
      tree->SetBranchAddress("digi.gem.hit.z", &(event->fHitZ) );
      SETBIT( fFlags, kHasHitpos );
    } else {
      // Ensure the hit position data are not left uninitialized
      memset( event->fHitX, 0, MC_MAXH*sizeof(Float_t) );
      memset( event->fHitY, 0, MC_MAXH*sizeof(Float_t) );
      memset( event->fHitZ, 0, MC_MAXH*sizeof(Float_t) );
      Warning( "SBSsimRun::Open", "No hit position data in input file %s",
	       rootFileName.Data() );
    }
    tree->SetBranchStatus("digi.gem.hit.mx", kTRUE);
    tree->SetBranchAddress("digi.gem.hit.mx", &(event->fHitMX) );
    tree->SetBranchStatus("digi.gem.hit.my", kTRUE);
    tree->SetBranchAddress("digi.gem.hit.my", &(event->fHitMY) );
    tree->SetBranchStatus("digi.gem.hit.mz", kTRUE);
    tree->SetBranchAddress("digi.gem.hit.mz", &(event->fHitMZ) );
    tree->SetBranchStatus("digi.gem.hit.charge", kTRUE);
    tree->SetBranchAddress("digi.gem.hit.charge", &(event->fCharge) );
    tree->SetBranchStatus("digi.gem.hit.time", kTRUE);
    tree->SetBranchAddress("digi.gem.hit.time", &(event->fTime) );
    tree->SetBranchStatus("digi.gem.hit.chamber", kTRUE);
    tree->SetBranchAddress("digi.gem.hit.chamber", &(event->fChamber) );
    tree->SetBranchStatus("digi.gem.hit.pID", kTRUE);
    tree->SetBranchAddress("digi.gem.hit.pID", &(event->fPID) );

    tree->SetBranchStatus("digi.gem.nch", kTRUE);
    tree->SetBranchAddress("digi.gem.nch", &(event->fNchan) );
    tree->SetBranchStatus("digi.gem.chamber", kTRUE);
    tree->SetBranchAddress("digi.gem.chamber", &(event->fPlane) );
    tree->SetBranchStatus("digi.gem.plane", kTRUE);
    tree->SetBranchAddress("digi.gem.plane", &(event->fAxis) );
    tree->SetBranchStatus("digi.gem.strip", kTRUE);
    tree->SetBranchAddress("digi.gem.strip", &(event->fStrip) );
    tree->SetBranchStatus("digi.gem.type", kTRUE);
    tree->SetBranchAddress("digi.gem.type", &(event->fType) );
    for( int k=0; k<MC_MAXSAMP; k++ ) {
      const char* var = Form("digi.gem.adc%d",k);
      tree->SetBranchStatus(var, kTRUE);
      tree->SetBranchAddress(var, &(event->fADC[k]) );
    }
  }

  // Kludge fake dates to be able to index into database
  if( TESTBIT(fFlags, kHasTracks) )
    // Assume this is g17 or later
    fDate.Set(20110302,160000);
  else if( TESTBIT(fFlags, kHasHitpos) )
    // Assume this is g16
    fDate.Set(20110208,0);
  else
    // Assume one of the early productions
    fDate.Set(20100401,0);

  nentries = tree->GetEntries();
  entry = 0;

  fOpened = kTRUE;
  return 0;
}

//-----------------------------------------------------------------------------
Int_t SBSsimRun::Close()
{
  delete tree; tree = 0;
  if (rootFile) {
    rootFile->Close();
    delete rootFile; rootFile = 0;
  }
  delete event; event = 0;
  fOpened = kFALSE;
  return 0;
}

//-----------------------------------------------------------------------------
Int_t SBSsimRun::ReadEvent()
{
  // Read next event from ROOT file

  if( entry >= nentries )
    return EOF;

  Int_t ret;
  if (!IsOpen()) {
    ret = Open();
    if (ret) return ret;
  }

  // Clear the event to get rid of anything still hanging around
  event->Clear();

  // Read input file
  ret = tree->GetEntry(entry++);
  if( ret == 0 )
    return EOF;
  if( ret < 0 )
    return -128;  // CODA_ERR

  // Extract the parameters of the trigger track
  SBSsimTrack* trk = new( (*event->fTracks)[0] ) SBSsimTrack();
  if( TESTBIT(fFlags, kHasTracks) ) {

    // Track position and momentum vectors in world coordinates
    TVector3 pos( event->fTrackX,  event->fTrackY,  event->fTrackZ  );
    TVector3 mom( event->fTrackMX, event->fTrackMY, event->fTrackMZ );

    // Transform vectors to spectrometer coordinates
    const Double_t angle = -16.9*TMath::DegToRad();
    pos.RotateY(angle);
    mom.RotateY(angle);
    
    // Convert to analyzer units (m, GeV)
    pos *= 1e-3;
    mom *= 1e-3;

    trk->fOrigin = pos;
    trk->fMomentum = mom;
    trk->fPID = 2212;
  } else {
    // Given Evaristo's hit-only information, getting the parameters of the
    // track is actually quite tricky. The horror!
    // 
    // The front of each chamber seems to be at -4.622 mm and no additional
    // hits ever occur until ~ -4.54 mm. So the first hit with z < -4.6
    // should be the first hit of the real track on the first chamber where
    // it actually registers a hit. Usually, that's the very first chamber,
    // but occasionally that chamber is simulated to be inefficient and the
    // hit is missing. Also, the index of this hit must obviously be < nsignal.
    // Finally, we need to select protons via the pID of the hit.
    Int_t i = 0;
    for( ; i < event->fNsignal; ++i ) {
      if( event->fPID[i] == 2212 && event->fHitMZ[i] > 1000. &&
	  (!TESTBIT(fFlags, kHasHitpos) || event->fHitZ[i] < -4.6) ) {

	trk->fMomentum.SetXYZ( event->fHitMX[i],
			       event->fHitMY[i],
			       event->fHitMZ[i] );
	trk->fMomentum *= 1e-3; // Convert to GeV

	// Set various track parameters
	trk->fPID = event->fPID[i];
	trk->fTimeOffset = event->fTime[i];

	// If this hit is not on the first chamber, the momentum in this way
	// will be slightly off due to energy loss and multiple scattering.
	// To be able to diagnose this effect later, we save the index of the
	// chamber where the parameters were extracted.
	trk->fHit1 = event->fChamber[i];
	assert( trk->fHit1 >= 0 );

	if( TESTBIT(fFlags, kHasHitpos) ) {
	  // If available, get the track position at the first chamber.
	  // The first chamber is assumed to be at z = 0 relative to the
	  // entire front tracker system (=detector coordinates)
	  if( trk->fHit1 == 0 ) {
	    trk->fOrigin.SetXYZ( event->fHitX[i], event->fHitY[i], 0. );
	  } else {
	    // Chambers are assumed to be 200 mm apart, evenly spaced.
	    // FIXME: Get chamber positions from SBSsimParameters
	    Float_t z  = trk->fHit1 * 200.;
	    Float_t mx = event->fHitMX[i] / event->fHitMZ[i];
	    Float_t my = event->fHitMY[i] / event->fHitMZ[i];
	    trk->fOrigin.SetXYZ( event->fHitX[i] - z*mx,
				 event->fHitY[i] - z*my,
				 0. );
	  }	  
	  trk->fOrigin *= 1e-3; // convert to meters
	}
	break; // found the track, so quit the loop
      }
    }
    // Nothing found!
    if( i == event->fNsignal ) {
      Warning( "SBSsimRun::ReadEvent", "No trigger track data found for event "
	       "number %d", event->fEvtNum );
    }
  }
  return S_SUCCESS;
}

//-----------------------------------------------------------------------------
const Int_t *SBSsimRun::GetEvBuffer() const
{
  if (!IsOpen()) return 0;

  return reinterpret_cast<Int_t*>(event);
}

//-----------------------------------------------------------------------------
SBSsimRun::~SBSsimRun()
{
  if (IsOpen())
    Close();
}

//_____________________________________________________________________________
void SBSsimRun::Print( Option_t* opt ) const
{
  // Print run info and status

  TString sopt(opt);
  bool do_header = sopt.Contains("start", TString::kIgnoreCase);
  if( sopt.IsNull() || do_header ) {
    THaRunBase::Print("STARTINFO");
  }

  if( sopt.IsNull() || sopt.Contains("status", TString::kIgnoreCase) ) {
    cout << "Analyzed events:       " << fNumAnalyzed << endl;
    cout << "Initialized:           " << fIsInit << endl;
    cout << "Opened:                " << fOpened << endl;
  }

  if( sopt.IsNull() || !do_header ) {

  //  if( fParam )
  //    fParam->Print(opt);
  }
}

//-----------------------------------------------------------------------------
ClassImp(SBSsimRun)
