#ifndef __CINT__

#define TESTCODE
#define DEBUG
#define VERBOSE

#include "GEM.h"
#include "SBS.h"
#include "SBSsim/SBSsim.h"
#include "SBSsim/SBSsimDecoder.h"
#include "SBSsim/SBSsimRun.h"
#include "TSystem.h"
#include "THaInterface.h"
#include "TList.h"
#include "TString.h"
#include "TFile.h"
#include "TH1I.h"
#include "TTree.h"
#include "TCut.h"
#include "TPad.h"
#include "THaGlobals.h"
#include "THaTextvars.h"
#include "THaAnalyzer.h"
#include <string>
#include <iostream>
#include <vector>
#include <algorithm>
#include <functional>

using namespace std;
using namespace TreeSearch;

struct PrintSBSsimRun :
  public std::unary_function< SBSsimRun*, void >
{
  void operator() ( const SBSsimRun* a ) const
  {
    cout << a->GetFileName() << "\t" << a->GetTitle() << endl;
  }
};
#endif

#define ALL(c) (c).begin(), (c).end()

void replay_xy() {

  THaInterface::SetDecoder( SBSsimDecoder::Class() );

  THaApparatus* SBS = new TreeSearch::SBS( "SBS", "SBS apparatus" );
  THaTrackingDetector* gem = 
    new TreeSearch::GEM( "gem_xy", "GEM trackers with xy-only readouts" );
  SBS->AddDetector( gem );
  gHaApps->Add( SBS );

  SBS->SetDebugAll(1);
  //  gem->THaTrackingDetector::SetDebug(2);

  TString db_prefix = SBS->GetName();
  db_prefix += ".";
  db_prefix += gem->GetName();
  gHaTextvars->Add( "DET", db_prefix.Data() );
  gHaTextvars->Add( "SBS", SBS->GetName() );

  THaAnalyzer* analyzer = THaAnalyzer::GetInstance();
  if( analyzer )
    analyzer->Close();
  else
    analyzer = new THaAnalyzer;

  TString odeffile("SBSsim_xy.odef");
  analyzer->EnableBenchmarks();
  analyzer->SetOdefFile(odeffile);
  analyzer->SetCrateMapFileName("SBSsim/db_SBSsim_cratemap.dat");
  analyzer->SetCountMode(THaAnalyzer::kCountPhysics);

  bool good = false;
  string inp;
  cout << "Production? " << flush;
  cin >> inp;
  string dir = "MC/" + inp + "/";
  void* dirp = gSystem->OpenDirectory(dir.c_str());
  vector<string> fnames;
  vector<SBSsimRun*> runs;
  TString rootfile(Form("SBSsim_xy_%s.root",inp.c_str()));
  analyzer->SetOutFile(rootfile);
  if( dirp ) {
    const char* e;
    while( (e=gSystem->GetDirEntry(dirp)) ) {
      TString fn(e);
      if( !fn.EndsWith(".root") )
	continue;
      fn.Prepend(dir.c_str());
      fnames.push_back(string(fn.Data()));
    }
    gSystem->FreeDirectory(dirp);
    cout << "Found " << fnames.size() << " input files" << endl;
    sort( ALL(fnames) );
    for( UInt_t i=0; i<fnames.size(); ++i ) {
      SBSsimRun* run = new SBSsimRun( fnames[i].c_str(),
	    Form("%s production part %u",inp.c_str(),runs.size()) );
      run->Init();
      runs.push_back(run);
    }
    for_each( ALL(runs), PrintSBSsimRun() );
    cout << "Number of events? " << flush;
    Int_t iev;
    UInt_t nev;
    cin >> iev;
    nev = iev;
    for( UInt_t i=0; i<runs.size(); ++i ) {
      SBSsimRun* run = runs[i];
      run->SetLastEvent(nev);
      cout << "i = " << i << ", nev = " << nev << endl;
      Int_t ret = analyzer->Init(run);
      if( ret ) return;
      ret = analyzer->Process(run);
      cout << "analyzer ret = " << ret << endl;
      if( run->GetNumAnalyzed() >= nev )
	break;
      nev -= run->GetNumAnalyzed();
      cout << "nev left = " << nev << endl;
    }
    good = true;
  } else {
    cout << "No such directory: " << dir << endl;
  }
  
  analyzer->Close();
  for( UInt_t i=0; i<runs.size(); ++i ) {
    delete runs[i];
  }
  runs.clear();
  delete analyzer; analyzer = 0;
  gHaApps->Delete();

  if( good ) {
    TFile* f = new TFile(rootfile,"r");
    if( !f || f->IsZombie() )
      cerr << "Oops, can't open " << rootfile << endl;
  }
}

void plot_xy() {

  // Reconstruction uncertainties (from replay)
  const Float_t sigx    = 2.3e-5;
  const Float_t sigy    = 5e-5;   // est., not Gaussian
  const Float_t sigth   = 1.7e-4;
  const Float_t sigph   = 1.8e-4;

  // Number of sigmas within which we accept tracks
  const Float_t nsigma  = 3.0;

  // "Kinematics window"
  const Float_t x_kine  = 0.18;
  const Float_t y_kine  = 0.002;
  const Float_t th_kine = 0.03;
  const Float_t ph_kine = 0.0007;
  const Float_t zback   = 1.0;

  string inp;
  cout << "Production? " << flush;
  cin >> inp;

  if( inp.empty() || inp[0] != 'g' ) {
    cerr << "Invalid name, must be g#" << endl;
    return;
  }

  TString rootfile(Form("SBSsim_xy_%s.root",inp.c_str()));
  TFile* f = new TFile(rootfile,"r");
  if( !f or f->IsZombie() ) {
    cerr << "Can't open " << rootfile << endl;
    return;
  }

  TObject* obj = f->Get("T");
  if( !obj || obj->IsA() != TTree::Class() ) {
    cerr << "Can't find tree \"T\" in " << rootfile << endl;
    return;
  }
  TTree* T = static_cast<TTree*>(obj);

  Long64_t nevt = T->GetEntries();
  if( nevt == 0 ) {
    cerr << "No events in tree" << endl;
    return;
  }
  TH1* h_evt = new TH1I("h_evt","evtnum",5000,0.,5001.);
  TCut ctrk = Form("abs(SBS.tr.th)<%f&&abs(SBS.tr.ph)<%f"
		   "&&abs(SBS.tr.x+SBS.tr.th*%f)<%f"
		   "&&abs(SBS.tr.y+SBS.tr.ph*%f)<%f",
		   th_kine,ph_kine,zback,x_trk,zback,y_trk);
  TCut ckine = Form("abs(SBS.tr.x+SBS.tr.th*%f)<%f"
		    "&&abs(SBS.tr.y+SBS.tr.ph*%f)<%f"
		    "&&abs(SBS.tr.th)<%f"
		    "&&abs(SBS.tr.ph)<%f",
		    zback,x_kine,zback,y_kine,th_kine,ph_kine);
  TCut cntrk = "SBS.tr.n>0";
  T->Draw("Entry$>>h_evt", ctrk && cntrk, "goff" );

  Double_t c_trk=0, c_multi=0;
  for(Long64_t i=0; i<h_evt->GetNbinsX(); ++i) {
    if( h_evt->GetBinContent(i)>0.1 )
      c_trk += 1;
    if( h_evt->GetBinContent(i)>1.1 )
      c_multi += 1;
  }
  Double_t c_single = c_trk - c_multi;
  cout << "Effic = " << c_single/nevt*100. << "%" << endl;

  h_evt->Reset();
  c_trk = 0;
  T->Draw("Entry$>>h_evt", ckine && cntrk, "goff" );
  for(Long64_t i=0; i<h_evt->GetNbinsX(); ++i) {
    if( h_evt->GetBinContent(i)>0.1 &&
	h_evt->GetBinContent(i)<1.1 )
      c_trk += 1;
  }
  cout << "c_single = " << c_single << ", c_trk = " << c_trk << endl;
  cout << "Mis-reconstructed in kine window = " 
       << (c_trk-c_single)/c_trk*100.  << "%" << endl;

  T->Draw("SBS.tr.x:SBS.tr.y", ctrk && cntrk);
  //  gPad->SaveAs(Form("%s_trx_try.png",inp.c_str()));

  //  f->Close();
  //  delete f; f = 0;

}
